# coding=utf-8

"""Actual implementation of SSO"""


import base64
import enum
import hashlib
import hmac
import typing
from urllib.parse import parse_qs, urlencode

import logging
from django.conf import settings
from django.utils.crypto import constant_time_compare
from django.utils.encoding import force_bytes

SSOResponse = typing.NamedTuple(
    "SSOResponse",
    (
        ("nonce", bytes),
        ("redirect", str)
    )
)


SignedPayload = typing.NamedTuple(
    "SignedPayload",
    (
        ("sso", bytes),
        ("sig", bytes)
    )
)


class ConsumerType(enum.Enum):
    """
    Enum representing whether a costumer is a Discourse instance
    or something else.
    """

    DISCOURSE = "discourse"
    OTHER = "other"


class BaseSSOGenerator(object):

    """Base class for SSO generator."""

    def __init__(self, sso_key: typing.Union[str, bytes]):
        self.sso_key = force_bytes(sso_key)

    def _get_payload_singature(self, payload: typing.Union[str, bytes]) -> bytes:
        """Returns HMAC signature for ``payload``."""
        digest = hmac.new(self.sso_key, force_bytes(payload), digestmod=hashlib.sha256).hexdigest()
        return force_bytes(digest)

    def prepare_signed_payload(
            self,
            params: typing.Union[dict, typing.Sequence[typing.Tuple[str, str]]]
    ) -> SignedPayload:
        """Prepare payload and signature."""
        encoded_params = base64.b64encode(force_bytes(urlencode(params)))
        response_sig = self._get_payload_singature(encoded_params)
        return (
            ('sso', encoded_params),
            ('sig', response_sig),
        )

    def _verify_payload(
            self,
            payload: typing.Union[str, bytes],
            signature: typing.Union[str, bytes]
    ) -> bool:
        """Verifies that signature is correct for payload."""
        this_signature = self._get_payload_singature(payload)
        result = constant_time_compare(
            this_signature,
            force_bytes(signature)
        )
        return result

    @classmethod
    def create_url(cls, base_url, signed_payload) -> str:
        """Get full url with query string."""
        query_string = urlencode(signed_payload)
        return '%s?%s' % (base_url, query_string)


class SSOProducerUtils(BaseSSOGenerator):
    """Utils for SSO producer."""

    def __init__(
            self,
            sso_key: typing.Union[str, bytes],
            consumer_url: str,
            user: settings.AUTH_USER_MODEL,
            sso: typing.Union[str, bytes],
            sig: typing.Union[str, bytes]
    ):
        """
        Init function, gets unpacked parameters from SSO endpoint,
        as well as logged in user.
        """
        super().__init__(sso_key)
        self.consumer_url = consumer_url
        assert isinstance(self.consumer_url, str)
        self.user = user
        self.sso = force_bytes(sso)
        self.sig = force_bytes(sig)

    def validate(
            self,
            consumer_type: ConsumerType = ConsumerType.DISCOURSE
    ):
        """Validate this sso instance, this should perform sanity checks."""
        if consumer_type == ConsumerType.DISCOURSE:
            # Discourse produces SSO that ends with a new line
            # it you accidentally trim it you'll get invalid key,
            # and will not have a good time debugging it.
            if self.sso[-1:] != b'\n':
                raise ValueError(
                    "Invalid sso. SSO should end with new line, "
                    "(discourse generates sso string ending with "
                    "'\\n')."
                    "SSO IS: '{}'".format(self.sso)
                )

        payload = base64.b64decode(self.sso)
        if b'nonce' not in payload:
            raise ValueError(self.sso, payload)

    @property
    def request_payload(self) -> str:
        """Returns decoded request payload."""
        payload = base64.b64decode(self.sso)
        return payload

    def verify_signature(self) -> bool:
        """Verifies the signature."""
        return self._verify_payload(self.sso, self.sig)

    def get_nonce(self) -> bytes:
        """Returns nonce from request payload."""
        parsed_payload = parse_qs(self.request_payload)
        return parsed_payload[b'nonce'][0]

    def get_response_params(self) -> typing.Sequence[typing.Tuple[str, str]]:
        """
        Returns SSO response parameters.

        Tuple of tuples is returned instead of a dict to have
        deterministic ordering of parameters, as order of parameters
        changes signature, and makes testing harder.
        """
        return (
            ('nonce', self.get_nonce()),
            ('email', self.user.email),
            ('username', self.user.username),
            ('external_id', self.user.id),
        )

    def get_signed_payload(self) -> dict:
        """Returns signed GET parameters for redirect."""
        return self.prepare_signed_payload(self.get_response_params())

    def get_sso_redirect(self, signed_payload) -> str:
        """Get full sso redirect."""
        return self.create_url(self.consumer_url, signed_payload)


class SSOProviderService(object):
    """
    SSO provider service.
    """

    def __init__(self, sso_key):
        """
        :param sso_key: SSO shared secret key.
        """
        self.sso_key = sso_key

    def get_signed_url(
            self,
            user: settings.AUTH_USER_MODEL,
            sso: str,
            signature: str,
            redirect_to: str,
            consumer_type: ConsumerType = ConsumerType.DISCOURSE
    ) -> typing.Optional[str]:
        """
        Performs SSO, returning redirect url.

        :param user: User that will be logged in using SSO.
        :param sso: SSO string received from consumer
        :param signature: Signature for sso string
        :param redirect_to: Endpoint user will be redirected to
        :param consumer_type: type of SSO consumer, can be used to
                              customize SSO a bit.
        """

        gen = SSOProducerUtils(
            sso_key=self.sso_key,
            consumer_url=redirect_to,
            user=user,
            sso=sso,
            sig=signature
        )
        try:
            gen.validate(consumer_type=consumer_type)
        except ValueError:
            logging.exception("Invalid sso")
            return None
        if not gen.verify_signature():
            return None
        payload = gen.get_signed_payload()
        return gen.get_sso_redirect(payload)


class SSOClientUtils(BaseSSOGenerator):
    """
    Utils for clients to the SSO mechanism.
    """

    def __init__(
            self,
            sso_key: typing.Union[str, bytes],
            base_sso_url: str
    ):
        """
        Init function. Set sso key and sso url
        """
        super().__init__(sso_key)
        self.base_sso_url = base_sso_url

    def generate_sso_url(self, nonce: str, require_activation=False) -> str:
        """
        Generates url to sso with given nonce
        :param nonce: nonce to be used in the sso url
        :param require_activation: Set to false if email is not validated
        :return: Full url to sso provider with all required query parameters
        """
        payload = [
            ('nonce', nonce),
        ]
        if require_activation:
            payload.append(('require_activation', "true"))
        signed_payload = self.prepare_signed_payload(payload)
        return self.create_url(self.base_sso_url, signed_payload)

    def validate_sso_against_sid(self, sso: str, sig: str) -> bool:
        """

        :param sso: payload to be verified
        :param sig: signed payload
        :return: Return true if sso matches sig
        """
        return self._verify_payload(sso, sig)

    @classmethod
    def decode_client_data(cls, base64_payload: str) -> dict:
        """
        Decode payload to dict of param and values
        :param base64_payload: Payload encoded in base 64
        :return: Dict of parameters and values. Values are in array.
        """
        parsed_payload = parse_qs(base64.b64decode(base64_payload))
        return parsed_payload

    @classmethod
    def get_param(cls, sso_decoded_payload: dict, param_name: str) -> str:
        """
        Extract parameter from payload dict
        :param sso_decoded_payload: dict
        :param param_name: Name of the param
        :return: Value of the param or empty string
        """
        if param_name in sso_decoded_payload.keys():
            values = sso_decoded_payload[param_name]
            if values:
                return values[0].decode('ascii')
        return ''



"""
Fixtures for all tests.
Most of the values are copied from SSO example in discourse.
"""

import pytest

from django.contrib.auth import get_user_model


# pylint: disable=missing-docstring

@pytest.fixture()
def example_user():
    return get_user_model()(
        username="foo",
        email="foo@bar.pl",
        id=1
    )


@pytest.fixture()
def example_secret():
    return b'd836444a9e4084d5b224a60c208dce14'


@pytest.fixture()
def example_sso():
    return b'bm9uY2U9Y2I2ODI1MWVlZmI1MjExZTU4YzAwZmYxMzk1ZjBjMGI=\n'


@pytest.fixture()
def example_sig():
    return b'2828aa29899722b35a2f191d34ef9b3ce695e0e6eeec47deb46d588d70c7cb56'


@pytest.fixture()
def example_producer_url():
    return 'http://example.org/sso'


@pytest.fixture()
def example_sso_url():
    return 'http://www.example.com/discourse/sso'



"""Views module"""
import string
import logging

import random
from abc import ABC, abstractmethod
import redis

from django.contrib.auth import get_user_model, login
from django.contrib.auth.mixins import LoginRequiredMixin
from django.http.response import HttpResponseBadRequest, HttpResponseRedirect
from django.views.generic.base import View

from discourse_django_sso import utils
from discourse_django_sso.utils import ConsumerType


log = logging.getLogger(__name__)


class SSOProviderView(LoginRequiredMixin, View):
    """
    View that implements sso.
    """

    sso_secret = None
    sso_redirect = None
    consumer_type = ConsumerType.DISCOURSE

    def get(self, request, **kwargs):  # pylint: disable=unused-argument
        """Performs the SSO"""
        try:
            sso = request.GET['sso']
            sig = request.GET['sig']
        except KeyError:
            return HttpResponseBadRequest()
        redirect = utils.SSOProviderService(
            sso_key=self.sso_secret
        ).get_signed_url(
            user=request.user,
            redirect_to=self.sso_redirect,
            sso=sso,
            signature=sig,
            consumer_type=self.consumer_type
        )
        if redirect is None:
            return HttpResponseBadRequest()
        return HttpResponseRedirect(
            redirect_to=redirect
        )


class NonceService(ABC):
    """
    Service for managing nonce.
    """

    def __init__(self):
        self.rand = random.SystemRandom()

    @abstractmethod
    def generate_nonce(self):
        """
        Generate nonce.
        """
        val = ''.join(self.rand.choice(string.hexdigits) for _ in range(32))
        while self.is_nonce_already_generated(val):
            val = ''.join(self.rand.choice(string.hexdigits) for _ in range(32))  # pragma: no cover
        return val

    @abstractmethod
    def is_nonce_already_generated(self, nonce) -> bool:
        """
        Return true if nonce was already generated in the past
        :param nonce: Nonce to be checked
        """
        pass

    @abstractmethod
    def is_nonce_valid(self, nonce: str):
        """
        Returns true if nonce is valid
        :param nonce: Nonce to be checked
        """
        pass

    @abstractmethod
    def invalidate_nonce(self, nonce: str):
        """
        Invalidate nonce
        :param nonce: Nonce to be invalidated
        """
        pass


class InMemoryNonceService(NonceService):
    """
    In memory implementation of nonce service
    """

    def __init__(self):
        super().__init__()
        self.generated_nonces = set()
        self.invalid_nonces = set()

    def generate_nonce(self) -> str:
        val = super().generate_nonce()
        self.generated_nonces.add(val)
        return val

    def is_nonce_already_generated(self, nonce) -> bool:
        return nonce in self.generated_nonces or nonce in self.invalid_nonces

    def is_nonce_valid(self, nonce: str) -> bool:
        return nonce in self.generated_nonces and nonce not in self.invalid_nonces

    def invalidate_nonce(self, nonce: str):
        self.invalid_nonces.add(nonce)


class RedisNonceService(NonceService):  # pragma: no cover
    """
    Implementation of NonceService based on Redis
    """

    def __init__(self, host, port, password):
        super().__init__()
        self.host = host
        self.port = port
        self.password = password
        self.redis = None

    def connect(self):
        """
        Connect to Redis server
        """
        if self.redis is not None:
            return
        log.info('connecting to redis: %s', self.host)
        try:
            self.redis = redis.Redis(self.host, self.port, self.password)
        except Exception:
            log.error('Exception while connecting to redis')

    def is_nonce_already_generated(self, nonce) -> bool:
        self.connect()
        return self.redis.get(nonce) is not None

    def invalidate_nonce(self, nonce: str):
        self.connect()
        self.redis.set(nonce, 'INVALID')

    def generate_nonce(self):
        self.connect()
        val = super().generate_nonce()
        # TODO or maybe use timeout when setting value, so that it expires after some time:
        # setex(name, time, value)  # pylint: disable=wrong-spelling-in-comment
        # Set the value of key name to value that expires in time seconds. time can be represented
        # by an integer or a Python timedelta object.   # pylint: disable=wrong-spelling-in-comment
        self.redis.set(val, 'VALID')
        return val

    def is_nonce_valid(self, nonce: str):
        self.connect()
        val = self.redis.get(nonce)
        return val is not None and (val == 'VALID' or val == b'VALID')


class SSOClientView(View):
    """
    View for the client of the sso
    """
    sso_secret = None
    sso_url = None
    nonce_service = None

    def get(self, request, **kwargs):  # pylint: disable=unused-argument
        """
        View which generates redirect to sso provider.
        """
        if request.user.is_authenticated():
            return HttpResponseRedirect(
                redirect_to='/'
            )
        client_util = utils.SSOClientUtils(self.sso_secret, self.sso_url)
        nonce_val = self.nonce_service.generate_nonce()
        # TODO how do we know about email value - true/false
        sso_url = client_util.generate_sso_url(nonce_val, False)
        request.session['redirect_path'] = request.GET.get('next', '/')

        return HttpResponseRedirect(
            redirect_to=sso_url
        )


class SSOCreateSessionView(View):
    """
    View for processing of response from sso provider
    """

    sso_secret = None
    nonce_service = None

    def get(self, request, **kwargs):  # pylint: disable=unused-argument
        """Performs the session creation"""
        try:
            sso = request.GET['sso']
            sig = request.GET['sig']
        except KeyError:
            return HttpResponseBadRequest()
        client_util = utils.SSOClientUtils(self.sso_secret, None)

        if request.user.is_authenticated() or not client_util.validate_sso_against_sid(sso, sig):
            return HttpResponseBadRequest()

        user_data = client_util.decode_client_data(sso)
        nonce = client_util.get_param(user_data, b'nonce')
        if self.nonce_service.is_nonce_valid(nonce):
            try:
                # SSOClientView is storing redirect path in anonymous session
                redirect_path = request.session.get('redirect_path', '/')
                self.create_user_session(request,
                                         client_util.get_param(user_data, b'email'),
                                         client_util.get_param(user_data, b'external_id'),
                                         client_util.get_param(user_data, b'username'))
                return HttpResponseRedirect(
                    redirect_to=redirect_path
                )
            except ValueError:  # pragma: no cover
                # return bad request
                log.error('Error while creating user session')
            finally:
                self.nonce_service.invalidate_nonce(nonce)
        else:
            log.debug('SSOCreateSessionView invalid nonce for user %s', client_util.get_param(user_data, b'email'))

        return HttpResponseBadRequest()     # pragma: no cover

    @classmethod
    def create_user_session(cls, request, user_email, external_id, username):
        # pylint: disable=unused-argument
        """
        Create user session for given user. SSO validation was successful.
        :param user_email: User email
        :param external_id: User id, unique in the external system
        :param username: Optional value of user name
        :return:
        """
        if not user_email or not username:
            raise ValueError('Missing required field')  # pragma: no cover
        user_model = get_user_model()
        user = user_model.objects.get_or_create(
            username='$sso$' + username,
            email=user_email,
            # TODO what about password, should we set some random value
            defaults={'first_name': username, 'is_active': True})
        login(request, user[0])



"""discourse_django_sso_test_project URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/1.11/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  url(r'^$', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  url(r'^$', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.conf.urls import url, include
    2. Add a URL to urlpatterns:  url(r'^blog/', include('blog.urls'))
"""
from discourse_django_sso.views import InMemoryNonceService
from django.conf.urls import url
from django.contrib import admin
from discourse_django_sso import views
from discourse_django_sso.utils import ConsumerType
from discourse_django_sso_test_project import settings


class FixedInMemoryNonceService(InMemoryNonceService):
    def generate_nonce(self) -> str:
        val = "e15917e0320992596f638ebff7678202"
        self.generated_nonces.add(val)
        return val

nonce_service = FixedInMemoryNonceService()

urlpatterns = [
    url(r'^admin/', admin.site.urls),
    url(
        r'^sso/',
        views.SSOProviderView.as_view(
            sso_redirect=settings.DISCOURSE_SSO_REDIRECT,
            sso_secret=settings.DISCOURSE_SSO_KEY,
            consumer_type=ConsumerType.DISCOURSE
        ),
        name="sso"
    ),
    url(
        r'^client/',
        views.SSOClientView.as_view(
            sso_secret=settings.DISCOURSE_SSO_KEY,
            sso_url=settings.DISCOURSE_SSO_URL,
            nonce_service=nonce_service),
        name="client"
    ),
    url(
        r'^createSession/',
        views.SSOCreateSessionView.as_view(
            sso_secret=settings.DISCOURSE_SSO_KEY,
            nonce_service=nonce_service),
        name="createSession"
    )

]
